from django.core import serializers
from django.core.serializers.base import DeserializationError
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext

from reversion.errors import RevertError
from reversion.revisions import _get_options
from reversion.serializers import deserialize_instance, deserialize_raw_fields


def get_object_version(model, data, object_repr, format):
    version_options = _get_options(model)
    data = force_str(data.encode('utf8'))
    try:
        return deserialize_instance(format, data, use_natural_foreign_keys=version_options.use_natural_foreign_keys)
    except DeserializationError:
        raise RevertError(gettext('Could not load %(object_repr)s version - incompatible version data.') % {
            'object_repr': object_repr,
        })
    except serializers.SerializerDoesNotExist:
        raise RevertError(gettext('Could not load %(object_repr)s version - unknown serializer %(format)s.') % {
            'object_repr': object_repr,
            'format': format,
        })


def get_local_field_dict(model, object_version):
    """
    A dictionary mapping field names to field values in object version of the model.

    Parent links of inherited multi-table models will not be followed.
    """
    version_options = _get_options(model)
    obj = object_version.object
    field_dict = {}
    for field_name in version_options.fields:
        field = model._meta.get_field(field_name)
        if isinstance(field, models.ManyToManyField):
            # M2M fields with a custom through are not stored in m2m_data, but as a separate model.
            if object_version.m2m_data and field.attname in object_version.m2m_data:
                field_dict[field.attname] = object_version.m2m_data[field.attname]
        else:
            field_dict[field.attname] = getattr(obj, field.attname)
    return field_dict


def get_raw_field_dict(data, object_repr, format):
    try:
        return deserialize_raw_fields(format, data)
    except DeserializationError:
        raise RevertError(gettext('Could not load %(object_repr)s version - incompatible version data.') % {
            'object_repr': object_repr,
        })
    except serializers.SerializerDoesNotExist:
        raise RevertError(gettext('Could not load %(object_repr)s version - unknown serializer %(format)s.') % {
            'object_repr': object_repr,
            'format': format,
        })



from django.core.management.base import BaseCommand, CommandError

from pydjamodb.connection import TableConnection


class Command(BaseCommand):

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--noinput', '--no-input', action='store_false', dest='interactive',
            help='Tells Django to NOT prompt the user for input of any kind.',
        )

    def handle(self, **options):
        self.stdout.write('Init DynamoDB revisions')

        connection = TableConnection('reversion')
        if connection.exists_table():
            if options['interactive']:
                message = (
                    'This will delete existing revisions!\n'
                    'Are you sure you want to do this?\n\n'
                    "Type 'yes' to continue, or 'no' to cancel: "
                )
                if input(message) != 'yes':
                    raise CommandError('Init DynamoDB revisions cancelled.')
            connection.delete_table(wait=True)
        connection.create_table(
            **{
                'attribute_definitions': [
                    {'attribute_name': 'object_key', 'attribute_type': 'S'},
                    {'attribute_name': 'revision_id', 'attribute_type': 'S'},
                    {'attribute_name': 'date_created', 'attribute_type': 'S'},
                    {'attribute_name': 'object_content_type_key', 'attribute_type': 'S'},
                    {'attribute_name': 'is_removed', 'attribute_type': 'S'}
                ],
                'key_schema': [
                    {'key_type': 'RANGE', 'attribute_name': 'object_key'},
                    {'key_type': 'HASH', 'attribute_name': 'revision_id'}
                ],
                'global_secondary_indexes': [
                    {
                        'index_name': 'object_content_type_created_index',
                        'key_schema': [
                            {'AttributeName': 'date_created', 'KeyType': 'RANGE'},
                            {'AttributeName': 'object_content_type_key', 'KeyType': 'HASH'}
                        ],
                        'projection': {'ProjectionType': 'ALL'}
                    },
                    {
                        'index_name': 'object_content_type_key_removed_index',
                        'key_schema': [
                            {'AttributeName': 'is_removed', 'KeyType': 'RANGE'},
                            {'AttributeName': 'object_content_type_key', 'KeyType': 'HASH'}
                        ],
                        'projection': {'ProjectionType': 'ALL'}
                    },
                    {
                        'index_name': 'object_date_created_index',
                        'key_schema': [
                            {'AttributeName': 'date_created', 'KeyType': 'RANGE'},
                            {'AttributeName': 'object_key', 'KeyType': 'HASH'}
                        ],
                        'projection': {'ProjectionType': 'ALL'}
                    }
                ],
                'local_secondary_indexes': []
            },
            wait=True
        )



from django.conf import settings
from django.test.utils import override_settings
from test_app.models import TestModel
from test_app.tests.base import TestBase, TestModelMixin, LoginMixin


use_middleware = override_settings(
    MIDDLEWARE=settings.MIDDLEWARE + ["reversion.middleware.RevisionMiddleware"],
)


@use_middleware
class RevisionMiddlewareTest(TestModelMixin, TestBase):

    def testCreateRevision(self):
        response = self.client.post("/test-app/save-obj/")
        obj = TestModel.objects.get(pk=response.content)
        self.assertSingleRevision((obj,), comment='Request log from "RevisionMiddleware", path "/test-app/save-obj/"')

    def testCreateRevisionError(self):
        with self.assertRaises(Exception):
            self.client.post("/test-app/save-obj-error/")
        self.assertNoRevision()


@use_middleware
class RevisionMiddlewareUserTest(TestModelMixin, LoginMixin, TestBase):

    def testCreateRevisionUser(self):
        response = self.client.post("/test-app/save-obj/")
        obj = TestModel.objects.get(pk=response.content)
        self.assertSingleRevision(
            (obj,), user=self.user, comment='Request log from "RevisionMiddleware", path "/test-app/save-obj/"'
        )

    @override_settings(REVERSION_ENABLED=False)
    def testDisabledRevision(self):
        response = self.client.post("/test-app/save-obj/")
        TestModel.objects.get(pk=response.content)
        self.assertNoRevision()


import import_string

from django.conf import settings


BACKENDS = [
    app_name.split('.')[-1] for app_name in settings.INSTALLED_APPS
    if isinstance(app_name, str) and app_name.startswith('reversion.backends.')
]

MODELS = {
    backend: import_string('reversion.backends.{}.models'.format(backend)) for backend in BACKENDS
}


def prepare_version_object(obj, content_type, object_id, model_db, version_options, explicit, using, is_delete):
    return MODELS[getattr(settings, 'REVERSION_BACKEND', None) or BACKENDS[0]].prepare_version_object(
        obj, content_type, object_id, model_db, version_options, explicit, using, is_delete
    )


def save_revision(date_created, user, comment, versions, using):
    return MODELS[getattr(settings, 'REVERSION_BACKEND', None) or BACKENDS[0]].save_revision(
        date_created, user, comment, versions, using
    )


def get_db_name():
    return MODELS[getattr(settings, 'REVERSION_BACKEND', None) or BACKENDS[0]].get_db_name()


def get_revision_or_none(id):
    return MODELS[getattr(settings, 'REVERSION_BACKEND', None) or BACKENDS[0]].get_revision_or_none(id)


if len(BACKENDS) == 1:
    Version = import_string('reversion.backends.{}.models.Version'.format(BACKENDS[0]))
    Reversion = import_string('reversion.backends.{}.models.Version'.format(BACKENDS[0]))



import import_string

from django.core.exceptions import ImproperlyConfigured
from django.core.serializers.base import SerializerDoesNotExist
from django.core import serializers


SERIALIZERS = (
    'reversion.serializers.json.JsonSerializer',
)


_serializers = {}


def get_serializer_formats():
    if not _serializers:
        _load_serializers()
    return list(_serializers)


def get_serializer(format):
    if not _serializers:
        _load_serializers()
    if format not in _serializers:
        raise SerializerDoesNotExist(format)
    return _serializers[format]


def serialize_instance(format, instance, **options):
    return get_serializer(format).serialize_instance(instance, **options)


def deserialize_instance(format, data, **options):
    return get_serializer(format).deserialize_instance(data, **options)


def deserialize_raw_fields(format, data, **options):
    return get_serializer(format).deserialize_raw_fields(data, **options)


def _load_serializers():
    for serializer_class_str in SERIALIZERS:
        try:
            serializer_class = import_string(serializer_class_str)
            _serializers[serializer_class.format] = serializer_class()
        except ImportError:
            raise ImproperlyConfigured(f'Missing reversion serializer with on path {serializer_class_str}')


class BaseSerializer:

    format = None

    def serialize_instance(self, instance, **options):
        return serializers.serialize(self.format, (instance,), **options)

    def deserialize_instance(self, data, **options):
        return list(serializers.deserialize(self.format, data, ignorenonexistent=True, **options))[0]

    def _deserialize_raw(self, data):
        raise NotImplementedError

    def deserialize_raw_fields(self, data):
        return self._deserialize_raw(data)[0]['fields']



from itertools import zip_longest

from pydjamodb.queryset import DynamoDBQuerySet


NULL_OBJ_KEY = '-'


class ObjectVersionDynamoDBQuerySet(DynamoDBQuerySet):

    def __init__(self, model):
        super().__init__(model)
        self._index = model.object_date_created_index
        self._prefetch_prev_versions = False
        self._scan_index_forward = False

    def _clone(self):
        c = super()._clone()
        c._prefetch_prev_versions = self._prefetch_prev_versions
        return c

    def set_index(self, index):
        raise RuntimeError('Value cannot be set')

    def prefetch_prev_versions(self):
        obj = self._clone()
        obj._prefetch_prev_versions = True
        return obj

    def _process_execution_with_prefetch_prev_version(self):
        """
        Execution prefetch prev version objects to loaded version.
        Because versions are ordered by creation date in queryset the result can be done with one or two DB requests.

        _scan_index_forward==True
        prev version of first object cannot be loaded from DB with this queryset
        and will not be preloaded (will be loaded with the lazy mannear of property prev_version in extra DB request)

         _scan_index_forward==True
         prev versions of all objects are loaded with one request. If limit is set it will be increased about 1 to
         get prev version for the last object and finally decreased back. If limit is not set the situation is much
         simplier because the last object has no prev version.
        """
        if self._scan_index_forward:
            super()._process_execution()

            reverse_results = self._results[::-1]
            self._prev_versions = reverse_results[1:]
            for version, prev_version in zip(reverse_results, self._prev_versions):
                version.prev_version = prev_version
        else:
            if self._limit:
                self._limit += 1
                super()._process_execution()
                self._limit -= 1

                results = self._results

                self._prev_versions = results[1:]
                # there are more results than limit
                self._results = results[:self._limit]
                for version, prev_version in zip_longest(self._results, self._prev_versions):
                    version.prev_version = prev_version
                if len(results) > self._limit:
                    self._next_key = {
                        key: self._results[-1].serialize()[key]
                        for key in self._execution.page_iter.key_names
                    }
            else:
                super()._process_execution()
                self._prev_versions = self._results[1:]
                for version, prev_version in zip_longest(self._results, self._prev_versions):
                    version.prev_version = prev_version

    def _process_execution(self):
        if self._prefetch_prev_versions:
            self._process_execution_with_prefetch_prev_version()
        else:
            super()._process_execution()

    def get_for_object_reference(self, model, object_id, model_db=None):
        from .models import get_key_from_content_type_and_id, _get_content_type

        return self.set_hash_key(
            get_key_from_content_type_and_id(_get_content_type(model), object_id, model_db)
        )

    def get_for_object(self, obj, model_db=None):
        return self.get_for_object_reference(obj.__class__, obj.pk, model_db=model_db)

    def get_for_model(self, model, model_db=None):
        from .models import get_object_content_type_key, _get_content_type

        return self._model.objects_all.set_index(self._model.object_content_type_created_index).set_hash_key(
            get_object_content_type_key(_get_content_type(model), model_db)
        )

    def get_deleted(self, model, model_db=None):
        from .models import get_object_content_type_key, _get_content_type

        return self._model.objects_all.set_index(self._model.object_content_type_key_removed_index).set_hash_key(
            get_object_content_type_key(_get_content_type(model), model_db)
        )


class RevisionDynamoDBQuerySet(DynamoDBQuerySet):

    def __init__(self, model):
        super().__init__(model)
        self._filter = self._get_filter(self._get_field('object_key'), 'eq',  NULL_OBJ_KEY)

    def all(self):
        return self.__class__(self._model)


class ObjectVersionRevisionDynamoDBQuerySet(DynamoDBQuerySet):

    def __init__(self, model):
        super().__init__(model)
        self._index = model.object_date_created_index
        self._prefetch_prev_versions = False
        self._hash_key = NULL_OBJ_KEY

    def all(self):
        return self.__class__(self._model)



from collections import namedtuple, defaultdict
from contextlib import contextmanager
from functools import wraps
from threading import local
from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.db import models, transaction, router
from django.db.models.query import QuerySet
from django.db.models.signals import post_save, m2m_changed, post_delete
from django.utils.encoding import force_str
from django.utils import timezone
from reversion.errors import RevisionManagementError, RegistrationError
from reversion.signals import pre_revision_commit, post_revision_commit


__all__ = (
    'pre_revision_commit',
    'post_revision_commit',
)


_VersionOptions = namedtuple("VersionOptions", (
    "fields",
    "follow",
    "format",
    "for_concrete_model",
    "ignore_duplicates",
    "use_natural_foreign_keys",
))


_StackFrame = namedtuple("StackFrame", (
    "manage_manually",
    "user",
    "comment",
    "date_created",
    "db_versions",
    "meta",
))


class _Local(local):

    def __init__(self):
        self.stack = ()


_local = _Local()


def is_active():
    return bool(_local.stack)


def deactivate():
    while is_active():
        _pop_frame()


def _current_frame():
    if not is_active():
        raise RevisionManagementError("There is no active revision for this thread")
    return _local.stack[-1]


def _copy_db_versions(db_versions):
    return {
        db: versions.copy()
        for db, versions
        in db_versions.items()
    }


def _push_frame(manage_manually, using):
    if is_active():
        current_frame = _current_frame()
        db_versions = _copy_db_versions(current_frame.db_versions)
        db_versions.setdefault(using, {})
        stack_frame = current_frame._replace(
            manage_manually=manage_manually,
            db_versions=db_versions,
        )
    else:
        stack_frame = _StackFrame(
            manage_manually=manage_manually,
            user=None,
            comment="",
            date_created=timezone.now(),
            db_versions={using: {}},
            meta=(),
        )
    _local.stack += (stack_frame,)


def _update_frame(**kwargs):
    _local.stack = _local.stack[:-1] + (_current_frame()._replace(**kwargs),)


def _pop_frame():
    prev_frame = _current_frame()
    _local.stack = _local.stack[:-1]
    if is_active():
        current_frame = _current_frame()
        db_versions = {
            db: prev_frame.db_versions[db]
            for db
            in current_frame.db_versions.keys()
        }
        _update_frame(
            user=prev_frame.user,
            comment=prev_frame.comment,
            date_created=prev_frame.date_created,
            db_versions=db_versions,
            meta=prev_frame.meta,
        )


def is_manage_manually():
    if getattr(settings, 'REVERSION_ENABLED', True):
        return _current_frame().manage_manually
    else:
        return False


def set_user(user):
    if getattr(settings, 'REVERSION_ENABLED', True):
        _update_frame(user=user)


def get_user():
    if getattr(settings, 'REVERSION_ENABLED', True):
        return _current_frame().user
    else:
        return None


def set_comment(comment):
    if getattr(settings, 'REVERSION_ENABLED', True):
        _update_frame(comment=comment)


def get_comment():
    if getattr(settings, 'REVERSION_ENABLED', True):
        return _current_frame().comment
    else:
        return None


def set_date_created(date_created):
    if getattr(settings, 'REVERSION_ENABLED', True):
        _update_frame(date_created=date_created)


def get_date_created():
    if getattr(settings, 'REVERSION_ENABLED', True):
        return _current_frame().date_created
    else:
        return None


def add_meta(model, **values):
    if getattr(settings, 'REVERSION_ENABLED', True):
        _update_frame(meta=_current_frame().meta + ((model, values),))


def _follow_relations(obj):
    version_options = _get_options(obj.__class__)
    for follow_name in version_options.follow:
        try:
            follow_obj = getattr(obj, follow_name)
        except ObjectDoesNotExist:
            continue
        if isinstance(follow_obj, models.Model):
            yield follow_obj
        elif isinstance(follow_obj, (models.Manager, QuerySet)):
            for follow_obj_instance in follow_obj.all():
                yield follow_obj_instance
        elif follow_obj is not None:
            raise RegistrationError("{name}.{follow_name} should be a Model or QuerySet".format(
                name=obj.__class__.__name__,
                follow_name=follow_name,
            ))


def _follow_relations_recursive(obj):
    def do_follow(obj):
        if obj not in relations:
            relations.add(obj)
            for related in _follow_relations(obj):
                do_follow(related)
    relations = set()
    do_follow(obj)
    return relations


def _add_to_revision(obj, using, model_db, explicit, is_delete):
    from reversion.models import prepare_version_object
    # Exit early if the object is not fully-formed.
    if obj.pk is None:
        return
    version_options = _get_options(obj.__class__)
    content_type = _get_content_type(obj.__class__, using)
    object_id = force_str(obj.pk)
    version_key = (content_type, object_id)
    # If the obj is already in the revision, stop now.
    db_versions = _current_frame().db_versions
    versions = db_versions[using]
    if version_key in versions and not explicit:
        return

    version = prepare_version_object(
        obj, content_type, object_id, model_db, version_options, explicit, using, is_delete
    )

    # If the version is a duplicate, stop now.
    if not version:
        return

    # Store the version.
    db_versions = _copy_db_versions(db_versions)
    db_versions[using][version_key] = version
    _update_frame(db_versions=db_versions)
    # Follow relations.
    for follow_obj in _follow_relations(obj):
        _add_to_revision(follow_obj, using, model_db, False, is_delete)


def add_to_revision(obj, model_db=None, is_delete=False):
    model_db = model_db or router.db_for_write(obj.__class__, instance=obj)
    for db in _current_frame().db_versions.keys():
        _add_to_revision(obj, db, model_db, True, is_delete)


def _save_revision(versions, user=None, comment="", meta=(), date_created=None, using=None):
    from reversion.models import save_revision
    # Only save versions that exist in the database.
    # Use _base_manager so we don't have problems when _default_manager is overriden
    model_db_pks = defaultdict(lambda: defaultdict(set))
    for version in versions:
        model_db_pks[version._model][version.db].add(version.object_id)
    model_db_existing_pks = {
        model: {
            db: frozenset(map(
                force_str,
                model._base_manager.using(db).filter(pk__in=pks).values_list("pk", flat=True),
            ))
            for db, pks in db_pks.items()
        }
        for model, db_pks in model_db_pks.items()
    }
    versions = [
        version for version in versions
        if version.is_delete or version.object_id in model_db_existing_pks[version._model][version.db]
    ]
    # Bail early if there are no objects to save.
    if not versions:
        return

    # Save a new revision.
    revision_id = save_revision(date_created, user, comment, versions, using)

    # Save the meta information.
    for meta_model, meta_fields in meta:
        meta_model._base_manager.db_manager(using=using).create(
            revision_id=revision_id,
            **meta_fields
        )


@contextmanager
def _dummy_context():
    yield


@contextmanager
def _create_revision_context(manage_manually, using, atomic, middleware):
    if getattr(settings, 'REVERSION_ENABLED', True):
        context = transaction.atomic(using=using) if atomic and using else _dummy_context()
        with context:
            _push_frame(manage_manually, using)
            try:
                try:
                    yield
                    # Only save for a db if that's the last stack frame for that db.
                    if not any(using in frame.db_versions for frame in _local.stack[:-1]):
                        current_frame = _current_frame()
                        _save_revision(
                            versions=current_frame.db_versions[using].values(),
                            user=current_frame.user,
                            comment=current_frame.comment,
                            meta=current_frame.meta,
                            date_created=current_frame.date_created,
                            using=using,
                        )
                finally:
                    _pop_frame()
            except RevisionManagementError:
                if not middleware:
                    raise
    else:
        yield


def create_revision(manage_manually=False, using=None, atomic=True, middleware=False):
    from reversion.models import get_db_name

    return _ContextWrapper(_create_revision_context, (manage_manually, using or get_db_name(), atomic, middleware))


class _ContextWrapper(object):

    def __init__(self, func, args):
        self._func = func
        self._args = args
        self._context = func(*args)

    def __enter__(self):
        return self._context.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return self._context.__exit__(exc_type, exc_value, traceback)

    def __call__(self, func):
        @wraps(func)
        def do_revision_context(*args, **kwargs):
            with self._func(*self._args):
                return func(*args, **kwargs)
        return do_revision_context


def _post_save_receiver(sender, instance, using, **kwargs):
    if is_registered(sender) and is_active() and not is_manage_manually():
        add_to_revision(instance, model_db=using)


def _post_delete_receiver(sender, instance, using, **kwargs):
    if is_registered(sender) and is_active() and not is_manage_manually():
        add_to_revision(instance, model_db=using, is_delete=True)


def _m2m_changed_receiver(instance, using, action, model, reverse, **kwargs):
    if action.startswith("post_") and not reverse:
        if is_registered(instance) and is_active() and not is_manage_manually():
            add_to_revision(instance, model_db=using)


def _get_registration_key(model):
    return (model._meta.app_label, model._meta.model_name)


_registered_models = {}


def is_registered(model):
    return _get_registration_key(model) in _registered_models


def get_registered_models():
    return (apps.get_model(*key) for key in _registered_models.keys())


def _get_senders_and_signals(model):
    yield model, post_save, _post_save_receiver
    yield model, post_delete, _post_delete_receiver
    opts = model._meta.concrete_model._meta
    for field in opts.local_many_to_many:
        m2m_model = field.remote_field.through
        if isinstance(m2m_model, str):
            if "." not in m2m_model:
                m2m_model = "{app_label}.{m2m_model}".format(
                    app_label=opts.app_label,
                    m2m_model=m2m_model
                )
        yield m2m_model, m2m_changed, _m2m_changed_receiver


def register(model=None, fields=None, exclude=(), follow=(), format="json",
             for_concrete_model=True, ignore_duplicates=False, use_natural_foreign_keys=False):
    def register(model):
        # Prevent multiple registration.
        if is_registered(model):
            raise RegistrationError("{model} has already been registered with django-reversion".format(
                model=model,
            ))
        # Parse fields.
        opts = model._meta.concrete_model._meta
        version_options = _VersionOptions(
            fields=tuple(
                field_name
                for field_name
                in ([
                    field.name
                    for field
                    in opts.local_fields + opts.local_many_to_many
                ] if fields is None else fields)
                if field_name not in exclude
            ),
            follow=tuple(follow),
            format=format,
            for_concrete_model=for_concrete_model,
            ignore_duplicates=ignore_duplicates,
            use_natural_foreign_keys=use_natural_foreign_keys,
        )
        # Register the model.
        _registered_models[_get_registration_key(model)] = version_options
        # Connect signals.
        for sender, signal, signal_receiver in _get_senders_and_signals(model):
            signal.connect(signal_receiver, sender=sender)
        # All done!
        return model
    # Return a class decorator if model is not given
    if model is None:
        return register
    # Register the model.
    return register(model)


def _assert_registered(model):
    if not is_registered(model):
        raise RegistrationError("{model} has not been registered with django-reversion".format(
            model=model,
        ))


def _get_options(model):
    _assert_registered(model)
    return _registered_models[_get_registration_key(model)]


def unregister(model):
    _assert_registered(model)
    del _registered_models[_get_registration_key(model)]
    # Disconnect signals.
    for sender, signal, signal_receiver in _get_senders_and_signals(model):
        signal.disconnect(signal_receiver, sender=sender)


def _get_content_type(model, using):
    from django.contrib.contenttypes.models import ContentType
    version_options = _get_options(model)
    return ContentType.objects.db_manager(using).get_for_model(
        model,
        for_concrete_model=version_options.for_concrete_model,
    )



from django.conf import settings

from reversion.revisions import create_revision, set_user, set_comment, deactivate


class RevisionMiddleware:

    """Wraps the entire request in a revision."""

    manage_manually = False

    using = None

    def __init__(self, get_response):
        self.get_response = get_response
        self.atomic = getattr(settings, 'REVERSION_ATOMIC_REVISION', True)

    def __call__(self, request):
        with create_revision(manage_manually=self.manage_manually, using=self.using, atomic=self.atomic,
                             middleware=True):
            if hasattr(request, 'user') and request.user is not None and request.user.is_authenticated:
                set_user(request.user)
            set_comment('Request log from "RevisionMiddleware", path "{}"'.format(request.path))
            response = self.get_response(request)
        deactivate()
        return response

    def process_exception(self, request, exception):
        """Closes the revision."""
        deactivate()



from pynamodb.attributes import UnicodeAttribute, UTCDateTimeAttribute
from pynamodb.indexes import AllProjection, GlobalSecondaryIndex

from pydjamodb.models import DynamoModel
from pydjamodb.attributes import BooleanUnicodeAttribute
from pydjamodb.queryset import DynamoDBManager

from uuid import uuid4

from django.core import serializers
from django.contrib.contenttypes.models import ContentType
from django.db import router
from django.utils.encoding import force_str
from django.utils.functional import cached_property

from reversion.backends.utils import get_object_version, get_local_field_dict, get_raw_field_dict
from reversion.revisions import _get_options
from reversion.signals import pre_revision_commit, post_revision_commit

from .queryset import (
    ObjectVersionDynamoDBQuerySet, RevisionDynamoDBQuerySet, ObjectVersionRevisionDynamoDBQuerySet, NULL_OBJ_KEY
)


def _get_content_type(model, using=None):
    version_options = _get_options(model)
    return ContentType.objects.db_manager(using).get_for_model(
        model,
        for_concrete_model=version_options.for_concrete_model,
    )


def get_key_from_content_type_and_id(content_type, object_id, model_db=None):
    model_db = model_db or router.db_for_write(content_type.model_class())
    return '|'.join((model_db, str(content_type.pk), str(object_id)))


def get_key_from_object(obj, model_db=None):
    if not obj:
        return NULL_OBJ_KEY
    else:
        return get_key_from_content_type_and_id(_get_content_type(obj), obj.pk, model_db)


def get_object_from_key_or_none(object_key):
    if not object_key or object_key == '-':
        return None

    model_db, content_type_id, object_id = object_key.split('|')

    return ContentType.objects.get(
        pk=content_type_id
    ).model_class().objects.using(model_db).filter(pk=object_id).first()


def get_object_content_type_key(content_type, model_db=None):
    model_db = model_db or router.db_for_write(content_type.model_class())
    return '{}|{}'.format(model_db, content_type.pk)


class VersionObjectDateCreatedIndex(GlobalSecondaryIndex):

    object_key = UnicodeAttribute(hash_key=True)
    date_created = UTCDateTimeAttribute(range_key=True)

    class Meta:
        projection = AllProjection()


class VersionModelDateCreatedIndex(GlobalSecondaryIndex):

    object_content_type_key = UnicodeAttribute(hash_key=True)
    date_created = UTCDateTimeAttribute(range_key=True)

    class Meta:
        projection = AllProjection()


class RemovedVersionIndex(GlobalSecondaryIndex):

    object_content_type_key = UnicodeAttribute(hash_key=True)
    is_removed = BooleanUnicodeAttribute(range_key=True)

    class Meta:
        projection = AllProjection()


class ReversionDynamoModel(DynamoModel):

    revision_id = UnicodeAttribute(hash_key=True)
    date_created = UTCDateTimeAttribute()
    user_key = UnicodeAttribute(null=True)
    comment = UnicodeAttribute(null=True)

    object_key = UnicodeAttribute(range_key=True, null=True)
    object_content_type_key = UnicodeAttribute(null=True)
    format = UnicodeAttribute(null=True)
    serialized_data = UnicodeAttribute(null=True)
    object_repr = UnicodeAttribute(null=True)
    is_removed = BooleanUnicodeAttribute(null=True)

    object_date_created_index = VersionObjectDateCreatedIndex()
    object_content_type_key_removed_index = RemovedVersionIndex()
    object_content_type_created_index = VersionModelDateCreatedIndex()

    @classmethod
    def from_raw_data(cls, data):
        """
        Returns an instance of this class
        from the raw data
        If object_key is equal to the NULL_OBJ_KEY value is instance ov Revision,
        otherwise Version instance is returned
        """
        if data is None:
            raise ValueError("Received no data to construct object")

        if data['object_key']['S'] == NULL_OBJ_KEY:
            return Revision._instantiate(data)
        else:
            return Version._instantiate(data)

    class Meta:
        table_name = 'reversion'

    @cached_property
    def user(self):
        return get_object_from_key_or_none(self.user_key)


class Revision(ReversionDynamoModel):

    objects = RevisionDynamoDBQuerySet.as_manager()
    objects_version = ObjectVersionRevisionDynamoDBQuerySet.as_manager()

    class Meta:
        proxy = True

    @property
    def version_set(self):
        return Version.objects_all.set_hash_key(self.revision_id).filter(object_key__startswith='VERSION')


class Version(ReversionDynamoModel):

    objects = ObjectVersionDynamoDBQuerySet.as_manager()
    objects_all = DynamoDBManager()

    class Meta:
        proxy = True

    @cached_property
    def revision(self):
        return Revision.get(self.revision_id, NULL_OBJ_KEY)

    @cached_property
    def _object_version(self):
        return get_object_version(self._model, self.serialized_data, self.object_repr, self.format)

    @cached_property
    def _local_field_dict(self):
        return get_local_field_dict(self._model, self._object_version)

    @cached_property
    def _local_raw_field_dict(self):
        return get_raw_field_dict(self.serialized_data, self.object_repr, self.format)

    def _get_parent_version_list(self):
        parent_version_list = []
        for parent_model, field in self._model._meta.concrete_model._meta.parents.items():
            parent_id = (
                self.object_id if self._model._meta.pk.attname == field.attname
                else self._local_field_dict.get(field.attname)
            )
            if parent_id:
                content_type = _get_content_type(parent_model)
                try:
                    parent_version_list.append(Version.get(
                        self.revision_id,
                        get_key_from_content_type_and_id(content_type, parent_id)
                    ))
                except Version.DoesNotExist:
                    pass
        return parent_version_list

    @cached_property
    def raw_field_dict(self):
        """
        A dictionary mapping field names to field values in this version
        of the model.
        This method will follow parent links, if present.
        """
        field_dict = self._local_raw_field_dict
        # Add parent data.
        for parent_version in self._get_parent_version_list():
            field_dict.update(parent_version.raw_field_dict)
        return field_dict

    @cached_property
    def field_dict(self):
        """
        A dictionary mapping field names to field values in this version
        of the model.
        This method will follow parent links, if present.
        """
        field_dict = self._local_field_dict
        # Add parent data.
        for parent_version in self._get_parent_version_list():
            field_dict.update(parent_version.field_dict)
        return field_dict

    @cached_property
    def object(self):
        return get_object_from_key_or_none(self.object_key)

    @property
    def object_id(self):
        return self.object_key.split('|')[2]

    @property
    def content_type_id(self):
        return int(self.object_key.split('|')[1])

    @property
    def _content_type(self):
        return ContentType.objects.get_for_id(self.content_type_id)

    @property
    def _model(self):
        return self._content_type.model_class()

    @cached_property
    def prev_version(self):
        return Version.objects.set_hash_key(self.object_key).filter(
            date_created__lt=self.date_created
        ).first()

    @property
    def is_revision(self):
        return not self.object_content_type_key

    @property
    def is_version(self):
        return not self.is_revision

    @property
    def db(self):
        return self.object_content_type_key.split('|')[0]

    @property
    def is_delete(self):
        return bool(self.is_removed)

    def revert(self):
        self._object_version.save(using=self.db)


def prepare_version_object(obj, content_type, object_id, model_db, version_options, explicit, using, is_delete):
    object_key = get_key_from_content_type_and_id(content_type, object_id, model_db)
    version = Version(
        object_key=object_key,
        format=version_options.format,
        serialized_data=serializers.serialize(
            version_options.format,
            (obj,),
            fields=version_options.fields,
            use_natural_foreign_keys=version_options.use_natural_foreign_keys,
        ),
        object_repr=force_str(obj),
        is_removed=True if is_delete else None,
        object_content_type_key=get_object_content_type_key(content_type, model_db)
    )

    if version_options.ignore_duplicates and explicit:
        previous_version = Version.objects.set_index(
            Version.object_date_created_index
        ).filter(object=object_key).first()
        if not is_delete and previous_version and previous_version._local_field_dict == version._local_field_dict:
            return None

    return version


def save_revision(date_created, user, comment, versions, using):
    from reversion.revisions import create_revision

    # Generate random revision PK
    revision_id = str(uuid4())

    user_key = get_key_from_object(user)

    revision = ReversionDynamoModel(
        revision_id=revision_id,
        object_key=NULL_OBJ_KEY,
        date_created=date_created,
        user_key=user_key,
        comment=comment
    )

    # Send the pre_revision_commit signal.
    pre_revision_commit.send(
        sender=create_revision,
        revision=revision,
        versions=versions
    )

    # Save version models.
    with Version.batch_write() as batch:
        batch.save(revision)

        for version in versions:
            version.revision_id = revision_id
            version.date_created = date_created
            version.user_key = user_key
            version.comment = comment
            batch.save(version)
    post_revision_commit.send(
        sender=create_revision,
        revision=revision,
        versions=versions,
        revision_id=revision_id
    )
    return revision_id


def get_db_name():
    return None


def get_revision_or_none(id):
    try:
        return Revision.get(id, NULL_OBJ_KEY)
    except Version.DoesNotExist:
        return None



from collections import defaultdict
from itertools import chain, groupby

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core import serializers
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, connections, models, router, transaction
from django.db.models.deletion import Collector
from django.db.models.functions import Cast
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from reversion.backends.utils import get_object_version, get_local_field_dict, get_raw_field_dict
from reversion.errors import RevertError
from reversion.revisions import _follow_relations_recursive, _get_content_type
from reversion.signals import pre_revision_commit, post_revision_commit


def _safe_revert(versions):
    unreverted_versions = []
    for version in versions:
        try:
            with transaction.atomic(using=version.db):
                version.revert()
        except (IntegrityError, ObjectDoesNotExist):
            unreverted_versions.append(version)
    if len(unreverted_versions) == len(versions):
        raise RevertError(gettext("Could not save %(object_repr)s version - missing dependency.") % {
            "object_repr": unreverted_versions[0],
        })
    if unreverted_versions:
        _safe_revert(unreverted_versions)


class Revision(models.Model):

    """A group of related serialized versions."""

    date_created = models.DateTimeField(
        db_index=True,
        verbose_name=_("date created"),
        help_text="The date and time this revision was created.",
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("user"),
        help_text="The user who created this revision.",
    )

    comment = models.TextField(
        blank=True,
        verbose_name=_("comment"),
        help_text="A text comment on this revision.",
    )

    def get_comment(self):
        try:
            LogEntry = apps.get_model('admin.LogEntry')
            return LogEntry(change_message=self.comment).get_change_message()
        except LookupError:
            return self.comment

    def revert(self, delete=False):
        # Group the models by the database of the serialized model.
        versions_by_db = defaultdict(list)
        for version in self.version_set.iterator():
            versions_by_db[version.db].append(version)
        # For each db, perform a separate atomic revert.
        for version_db, versions in versions_by_db.items():
            with transaction.atomic(using=version_db):
                # Optionally delete objects no longer in the current revision.
                if delete:
                    # Get a set of all objects in this revision.
                    old_revision = set()
                    for version in versions:
                        model = version._model
                        try:
                            # Load the model instance from the same DB as it was saved under.
                            old_revision.add(model._default_manager.using(version.db).get(pk=version.object_id))
                        except model.DoesNotExist:
                            pass
                    # Calculate the set of all objects that are in the revision now.
                    current_revision = chain.from_iterable(
                        _follow_relations_recursive(obj)
                        for obj in old_revision
                    )
                    # Delete objects that are no longer in the current revision.
                    collector = Collector(using=version_db)
                    new_objs = [item for item in current_revision
                                if item not in old_revision]
                    for model, group in groupby(new_objs, type):
                        collector.collect(list(group))
                    collector.delete()
                # Attempt to revert all revisions.
                _safe_revert(versions)

    def __str__(self):
        return ", ".join(force_str(version) for version in self.version_set.all())

    class Meta:
        ordering = ("-pk",)


class VersionQuerySet(models.QuerySet):

    def get_for_model(self, model, model_db=None):
        model_db = model_db or router.db_for_write(model)
        content_type = _get_content_type(model, self.db)
        return self.filter(
            content_type=content_type,
            db=model_db,
        )

    def get_for_object_reference(self, model, object_id, model_db=None):
        return self.get_for_model(model, model_db=model_db).filter(
            object_id=object_id,
        )

    def get_for_object(self, obj, model_db=None):
        return self.get_for_object_reference(obj.__class__, obj.pk, model_db=model_db)

    def get_deleted(self, model, model_db=None):
        model_db = model_db or router.db_for_write(model)
        connection = connections[self.db]
        if self.db == model_db and connection.vendor in ("sqlite", "postgresql", "oracle"):
            model_qs = (
                model._default_manager
                .using(model_db)
                .annotate(_pk_to_object_id=Cast("pk", Version._meta.get_field("object_id")))
                .filter(_pk_to_object_id=models.OuterRef("object_id"))
            )
            subquery = (
                self.get_for_model(model, model_db=model_db)
                .annotate(pk_not_exists=~models.Exists(model_qs))
                .filter(pk_not_exists=True)
                .values("object_id")
                .annotate(latest_pk=models.Max("pk"))
                .values("latest_pk")
            )
        else:
            # We have to use a slow subquery.
            subquery = self.get_for_model(model, model_db=model_db).exclude(
                object_id__in=list(
                    model._default_manager.using(model_db).values_list("pk", flat=True).order_by().iterator()
                ),
            ).values_list("object_id").annotate(
                latest_pk=models.Max("pk")
            ).order_by().values_list("latest_pk", flat=True)
        # Perform the subquery.
        return self.filter(pk__in=subquery)

    def get_unique(self):
        last_key = None
        for version in self.iterator():
            key = (version.object_id, version.content_type_id, version.db, version._local_field_dict)
            if last_key != key:
                yield version
            last_key = key


class Version(models.Model):

    """A saved version of a database model."""

    objects = VersionQuerySet.as_manager()

    revision = models.ForeignKey(
        Revision,
        on_delete=models.CASCADE,
        help_text="The revision that contains this version.",
    )

    object_id = models.CharField(
        max_length=191,
        help_text="Primary key of the model under version control.",
    )

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        help_text="Content type of the model under version control.",
    )

    is_delete = False

    @property
    def _content_type(self):
        return ContentType.objects.db_manager(self._state.db).get_for_id(self.content_type_id)

    @property
    def _model(self):
        return self._content_type.model_class()

    # A link to the current instance, not the version stored in this Version!
    object = GenericForeignKey(
        ct_field="content_type",
        fk_field="object_id",
    )

    db = models.CharField(
        max_length=191,
        help_text="The database the model under version control is stored in.",
    )

    format = models.CharField(
        max_length=255,
        help_text="The serialization format used by this model.",
    )

    serialized_data = models.TextField(
        help_text="The serialized form of this version of the model.",
    )

    object_repr = models.TextField(
        help_text="A string representation of the object.",
    )

    @cached_property
    def _object_version(self):
        return get_object_version(self._model, self.serialized_data, self.object_repr, self.format)

    @cached_property
    def _local_field_dict(self):
        return get_local_field_dict(self._model, self._object_version)

    @cached_property
    def _local_raw_field_dict(self):
        return get_raw_field_dict(self.serialized_data, self.object_repr, self.format)

    def _get_parent_version_list(self):
        field_dict = self._local_field_dict
        parent_version_list = []
        for parent_model, field in self._model._meta.concrete_model._meta.parents.items():
            content_type = _get_content_type(parent_model, self._state.db)
            parent_id = field_dict[field.attname]
            try:
                parent_version_list.append(
                    self.revision.version_set.get(
                        content_type=content_type,
                        object_id=parent_id,
                        db=self.db,
                    )
                )
            except Version.DoesNotExist:
                pass
        return parent_version_list

    @cached_property
    def raw_field_dict(self):
        """
        A dictionary mapping field names to field values in this version
        of the model.

        This method will follow parent links, if present.
        """
        field_dict = self._local_raw_field_dict
        # Add parent data.
        for parent_version in self._get_parent_version_list():
            field_dict.update(parent_version._local_raw_field_dict)
        return field_dict

    @cached_property
    def field_dict(self):
        """
        A dictionary mapping field names to field values in this version
        of the model.

        This method will follow parent links, if present.
        """
        field_dict = self._local_field_dict
        # Add parent data.
        for parent_version in self._get_parent_version_list():
            field_dict.update(parent_version._local_field_dict)
        return field_dict

    def revert(self):
        self._object_version.save(using=self.db)

    def __str__(self):
        return self.object_repr

    class Meta:
        unique_together = (
            ("db", "content_type", "object_id", "revision"),
        )
        ordering = ("-pk",)


class _Str(models.Func):

    """Casts a value to the database's text type."""

    function = "CAST"
    template = "%(function)s(%(expressions)s as %(db_type)s)"

    def __init__(self, expression):
        super().__init__(expression, output_field=models.TextField())

    def as_sql(self, compiler, connection):
        self.extra["db_type"] = self.output_field.db_type(connection)
        return super().as_sql(compiler, connection)


def _safe_subquery(method, left_query, left_field_name, right_subquery, right_field_name):
    right_subquery = right_subquery.order_by().values_list(right_field_name, flat=True)
    left_field = left_query.model._meta.get_field(left_field_name)
    right_field = right_subquery.model._meta.get_field(right_field_name)
    # If the databases don't match, we have to do it in-memory.
    # If it's not a supported database, we also have to do it in-memory.
    if (
        left_query.db != right_subquery.db or not
        (
            left_field.get_internal_type() != right_field.get_internal_type() and
            connections[left_query.db].vendor in ("sqlite", "postgresql")
        )
    ):
        return getattr(left_query, method)(**{
            "{}__in".format(left_field_name): list(right_subquery.iterator()),
        })
    else:
        # If the left hand side is not a text field, we need to cast it.
        if not isinstance(left_field, (models.CharField, models.TextField)):
            left_field_name_str = "{}_str".format(left_field_name)
            left_query = left_query.annotate(**{
                left_field_name_str: _Str(left_field_name),
            })
            left_field_name = left_field_name_str
        # If the right hand side is not a text field, we need to cast it.
        if not isinstance(right_field, (models.CharField, models.TextField)):
            right_field_name_str = "{}_str".format(right_field_name)
            right_subquery = right_subquery.annotate(**{
                right_field_name_str: _Str(right_field_name),
            }).values_list(right_field_name_str, flat=True)
            right_field_name = right_field_name_str
        # Use Exists if running on the same DB, it is much much faster
        exist_annotation_name = "{}_annotation_str".format(right_subquery.model._meta.db_table)
        right_subquery = right_subquery.filter(**{right_field_name: models.OuterRef(left_field_name)})
        left_query = left_query.annotate(**{exist_annotation_name: models.Exists(right_subquery)})
        return getattr(left_query, method)(**{exist_annotation_name: True})


def prepare_version_object(obj, content_type, object_id, model_db, version_options, explicit, using, is_delete):
    if is_delete:
        return None

    version = Version(
        content_type=content_type,
        object_id=object_id,
        db=model_db,
        format=version_options.format,
        serialized_data=serializers.serialize(
            version_options.format,
            (obj,),
            fields=version_options.fields,
            use_natural_foreign_keys=version_options.use_natural_foreign_keys,
        ),
        object_repr=force_str(obj),
    )
    if version_options.ignore_duplicates and explicit:
        previous_version = Version.objects.using(using).get_for_object(obj, model_db=model_db).first()
        if previous_version and previous_version._local_field_dict == version._local_field_dict:
            return None
    return version


def save_revision(date_created, user, comment, versions, using):
    from reversion.revisions import create_revision

    revision = Revision(
        date_created=date_created,
        user=user,
        comment=comment,
    )
    # Send the pre_revision_commit signal.
    pre_revision_commit.send(
        sender=create_revision,
        revision=revision,
        versions=versions,
    )
    # Save the revision.
    revision.save(using=using)
    # Save version models.
    for version in versions:
        version.revision = revision
        version.save(using=using)
    post_revision_commit.send(
        sender=create_revision,
        revision=revision,
        versions=versions,
        revision_id=revision.id
    )
    return revision.pk


def get_db_name():
    return router.db_for_write(Revision)


def get_revision_or_none(id):
    try:
        return Revision.objects.get(pk=id)
    except ObjectDoesNotExist:
        return None
